// njs_chisquare.cpp

#undef NJS_DEBUG

#include <njs_chisquare.h>

#include <njs_float.h>
#include <njs_function.h>
#include <njs_io.h>
#include <njs_rootfinder.h>
#include <njs_variate.h>

using namespace Njs;

double ChiSquare::rightTail (      // Chi Square right tail 
double df,
double chisquare,
int *error,
double rtol)
{
   if (chisquare <= 0.0) {
return 1.0;
   }
return *error == 0 ? Function::incompleteQGamma (0.5 * df, 0.5 * chisquare, error, rtol) : 0.0;
}

namespace {

   double s_chisquareDF = 0.0;
   int *s_error_ChiSquareRightTail = 0;
   double s_rtol_ChiSquareRightTail = 0.0;

}

static double l_chiSquareRightTail_s (double x)
{
   int error = 0;
   double y = ChiSquare::rightTail (s_chisquareDF, x, &error, s_rtol_ChiSquareRightTail);
   if (*s_error_ChiSquareRightTail == 0) {
      *s_error_ChiSquareRightTail = error;
   }
return y;
}

double ChiSquare::sigLevel (       // Chi Square significance level 
double df,                           // degrees of freedom 
double rightTail,                    // size of right tail
int *error,
double tol)
{
   if (*error != 0) {
return 0.0;
   }
	s_chisquareDF = df;
	s_error_ChiSquareRightTail = error;
   s_rtol_ChiSquareRightTail = tol * Function::RTOL; 
	if (rightTail <= 0.0) {
Njs::Io::abortFormat ("Function::chiSquareSigLevel : rightTail <= 0.0");
	} else if (1.0 < rightTail) {
Njs::Io::abortFormat ("Function::chiSquareSigLevel : 1.0 < rightTail");
	} else if (rightTail == 1.0) {
return 0.0;
	}
	double chisq = 0.0;
	double end = 0.0;
	for (end = 1.0; (chisq = l_chiSquareRightTail_s (end)) > rightTail; end += 1.0) {
      if (*error != 0) {
return Variate::BAD_PROBABILITY;
      }
   }
	if (chisq == rightTail) {
return end;
	}
return RootFinder::bisection (rightTail, l_chiSquareRightTail_s, end - 1.0, end, tol);
}

