// njs_function.h

#ifndef INCLUDED_NJS_FUNCTION
#define INCLUDED_NJS_FUNCTION

#include <njs_system.h>

/*****************************************************************************
 
*************************** FUNCTION ROUTINES  ***************************

*****************************************************************************/

namespace Njs {

   namespace Function {

      const double TOL =  1.0e-06;
      const double RTOL = 1.0e-06;

      double seriesSum (                          // series sum 
      int m,                                     // starting term 
      int n,                                     // ending term 
      void *p,                                    // parameters 
      double (*term_function) (int n, void *p),  // function giving term 
      int *error,                                // error status 
      double rtol = RTOL);                        // relative accuracy 

      double continuedFraction (                  // continued fraction limit 
      int m,                                     // starting term 
      int n,                                     // ending term 
      void *p,                                    // parameters 
      double (*a) (int n, void *p),              // function giving term 
      double (*b) (int n, void *p),              // function giving term 
      int *error,                                // error = 1 : premature termination of series
      double rtol = RTOL);                        // relative tolerance for error

      void setGammaCoeff                          // changes gamma coefficients 
	      (double *cof, int gammaSeries, double gamma);

      double factorial (double x);                // gamma(x+1) to 10 digits of accuracy 
	      
      double lnGamma (double x);                  // ln[gamma(x)] 

      double diGamma (double x);                  // digamma(x) 

      double triGamma (double x);                 // trigamma(x) 

      double incompleteQGamma (                   // Incomplete Gamma Function Q(a,x) 
      double a,
      double x,
      int *error,
      double rtol = RTOL);

      double erf                                  // Error Function 
	      (double x, int *error, double rtol = RTOL);

      double erfc                                 // Complementary Error Function 
	      (double x, int *error, double rtol = RTOL);

      double millsRatio                           // ratio of normal tail to normal density 
	      (double x, int *error, double rtol = RTOL);

      double oneMinusExp (double x);              // 1.0 - exp (x) 
	      
      template <class RealInt>
      double lnBinomial (RealInt n, RealInt r)    // natural log binomial coefficient 
      {
      return lnGamma ((double) (n + 1)) -
	          lnGamma ((double) (r + 1)) -
	          lnGamma ((double) (n - r + 1));
      }

      template <class Integer>
      double lnMultinomial (Integer n, Integer *r) // (double) natural log multinomial coefficient  
      {
	      Integer rsum = 0;
	      double ibico = 0.0;
	      for (Integer i = 0; i < n; i++) {
		      rsum += r[i];
		      ibico -= lnGamma ((double) (r [i] + 1));
	      }
	      ibico += lnGamma ((double) (rsum + 1));
      return ibico;
      }

   }

}

#endif //! INCLUDED_NJS_FUNCTION
